from functools import partial
from typing import (
    TYPE_CHECKING, Callable, Dict, List, Optional, Type, Union, cast, overload,
)

from dl_plus.exceptions import DLPlusException

from .peqn import PEQN


if TYPE_CHECKING:
    from .extractor import Extractor


EXTRACTOR_PEQN_ATTR = '_dl_plus_peqn'


ExtractorType = Type['Extractor']
RegisterDecoratorType = Callable[[ExtractorType], ExtractorType]


class ExtractorPluginError(DLPlusException):

    pass


class ExtractorPlugin:
    """
    An extractor plugin object

    :param str import_path: An import path of the plugin module.
        The :code:`__name__` module attribute should be used as a value:

        .. code-block::

            # dl_plus/extractors/somenamespace/someplugin.py
            plugin = ExtractorPlugin(__name__)

    :raises ExtractorPluginError: if the passed import path does not conform to
        plugin structure rules.
    """

    def __init__(self, import_path: str) -> None:
        try:
            self._base_peqn = PEQN.from_plugin_import_path(import_path)
        except ValueError as exc:
            raise ExtractorPluginError(exc)
        self._extractors: Dict[Optional[str], ExtractorType] = {}

    @overload
    def register(self, __extractor_cls: ExtractorType) -> ExtractorType:
        ...

    @overload
    def register(
        self, __extractor_cls: ExtractorType, *, name: str,
    ) -> ExtractorType:
        ...

    @overload
    def register(self, __name: str) -> RegisterDecoratorType:
        ...

    def register(
        self, __extractor_cls_or_name: Union[ExtractorType, str, None] = None,
        *, name: Optional[str] = None,
    ) -> Union[ExtractorType, RegisterDecoratorType]:
        """
        Register the given :class:`dl_plus.extractor.Extractor` class
        in the plugin's registry

        This method can be used as:

            * a regular method:

                .. code-block::

                    plugin.register(UnnamedExtractor)

                .. code-block::

                    plugin.register(NamedExtractor, name='clip')

            * a decorator:

                .. code-block::

                    @plugin.register
                    class UnnamedExtractor(Extractor):
                        ...

                .. code-block::

                    @plugin.register('clip')
                    class NamedExtractor(Extractor):
                        ...

        :raises ExtractorPluginError:
        """
        # A special case for `@plugin.extractor()` syntax
        if __extractor_cls_or_name is None:
            return partial(self._register, name=name)
        if isinstance(__extractor_cls_or_name, str):
            return partial(self._register, name=__extractor_cls_or_name)
        return self._register(__extractor_cls_or_name, name)

    def _register(
        self, extractor_cls: ExtractorType, name: Optional[str],
    ) -> ExtractorType:
        from .extractor import Extractor
        if not issubclass(extractor_cls, Extractor):
            raise ExtractorPluginError(
                f'Extractor subclass expected, got: {extractor_cls!r}')
        if EXTRACTOR_PEQN_ATTR in extractor_cls.__dict__:   # type: ignore
            peqn = cast(
                PEQN,
                extractor_cls.__dict__[EXTRACTOR_PEQN_ATTR],   # type: ignore
            )
            raise ExtractorPluginError(
                f'the extractor {extractor_cls!r} is already registered '
                f'as "{peqn}"'
            )
        if extractor_cls.IE_NAME is not None:
            raise ExtractorPluginError(
                f'the extractor {extractor_cls!r} has non-None IE_NAME value')
        if name and name in self._extractors:
            raise ExtractorPluginError(
                f'the plugin already contains an extractor called "{name}"')
        if name and None in self._extractors or not name and self._extractors:
            raise ExtractorPluginError(
                f'{extractor_cls!r}: the unnamed extractor must be '
                f'the only extractor in the plugin'
            )
        if 'DLP_REL_URL' in extractor_cls.__dict__:   # type: ignore
            rel_url = extractor_cls.DLP_REL_URL
            if not rel_url:
                raise ExtractorPluginError(
                    f'{extractor_cls!r}: DLP_REL_URL is not set')
            base_url = extractor_cls.DLP_BASE_URL
            if not base_url:
                raise ExtractorPluginError(
                    f'{extractor_cls!r}: DLP_REL_URL without DLP_BASE_URL')
            if '_VALID_URL' in extractor_cls.__dict__:   # type: ignore
                raise ExtractorPluginError(
                    f'{extractor_cls!r}: DLP_REL_URL and _VALID_URL '
                    f'are mutually exclusive'
                )
            extractor_cls._VALID_URL = base_url + rel_url
        peqn = self._base_peqn.copy(name=name)
        setattr(extractor_cls, EXTRACTOR_PEQN_ATTR, peqn)
        extractor_cls.IE_NAME = str(peqn)
        self._extractors[name or None] = extractor_cls
        return extractor_cls

    def get_extractor(self, name: Optional[str] = None) -> ExtractorType:
        """
        Get the extractor class from the plugin's registry

        :param str name: (optional) The name of the extractor. The same value
            that was used to register the extractor (see :meth:`register`).
            Omit this parameter to get the unnamed extractor.
        :rtype: dl_plus.extractor.Extractor
        :raises KeyError: if no extractor found
        """
        return self._extractors[name]

    def get_all_extractors(self) -> List[ExtractorType]:
        """
        Get all extractors from the plugin's registry

        :rtype: list[dl_plus.extractor.Extractor]
        """
        return list(self._extractors.values())



from typing import Optional, Union


class DLPlusException(Exception):

    def __init__(self, message: Union[str, Exception, None] = None) -> None:
        if message is not None:
            message = str(message)
        self._message = message

    def __str__(self):
        if self._message is None:
            if self.__cause__:
                self._message = str(self.__cause__)
            else:
                self._message = ''
        return self._message

    @property
    def error(self) -> Optional[BaseException]:
        return self.__cause__



import re
from typing import Optional, Union

from dl_plus.const import PLUGINS_PACKAGE
from dl_plus.utils import NOTSET, NotSet


PLUGIN_IMPORT_PATH_PREFIX = f'{PLUGINS_PACKAGE}.'
PART_REGEX = re.compile(r'^[a-z0-9]+$')
PEQN_REGEX = re.compile(
    r'^(?P<ns>[a-z0-9]+)/(?P<plugin>[a-z0-9]+)(?::(?P<name>[a-z0-9]+))?$')


def _check_part(part: str, part_name: str) -> None:
    if not part:
        raise ValueError(f'empty {part_name} part')
    if not PART_REGEX.fullmatch(part):
        raise ValueError(f'bad {part_name} part: {part}')


def _escape_import_part(part: str) -> str:
    try:
        if part[0].isdigit():
            return '_' + part
    except IndexError:
        pass
    return part


def _unescape_import_part(part: str) -> str:
    try:
        if part[0] == '_':
            return part[1:]
    except IndexError:
        pass
    return part


class PEQN:
    """
    Pluggable Extractor Qualified Name
    """

    def __init__(
        self, ns: str, plugin: str, name: Optional[str] = None,
    ) -> None:
        _check_part(ns, 'ns')
        _check_part(plugin, 'plugin')
        if name is not None:
            _check_part(name, 'name')
        self._ns = ns
        self._plugin = plugin
        self._name = name

    @classmethod
    def from_plugin_import_path(cls, path: str) -> 'PEQN':
        try:
            if not path.startswith(PLUGIN_IMPORT_PATH_PREFIX):
                raise ValueError('not in plugins package')
            parts = path[len(PLUGIN_IMPORT_PATH_PREFIX):].split('.')
            if len(parts) < 2:
                raise ValueError('not enough parts')
            elif len(parts) > 2:
                raise ValueError('too many parts')
            ns = _unescape_import_part(parts[0])
            plugin = _unescape_import_part(parts[1])
            return cls(ns, plugin)
        except ValueError as exc:
            raise ValueError(f'bad plugin import path: {path}: {exc}')

    @classmethod
    def from_string(cls, peqn: str) -> 'PEQN':
        match = PEQN_REGEX.fullmatch(peqn)
        if not match:
            raise ValueError(f'bad PEQN: {peqn}')
        return cls(**match.groupdict())

    def copy(
        self,
        ns: Union[str, NotSet] = NOTSET,
        plugin: Union[str, NotSet] = NOTSET,
        name: Union[str, None, NotSet] = NOTSET,
    ) -> 'PEQN':
        if ns is NOTSET:
            ns = self._ns
        if plugin is NOTSET:
            plugin = self._plugin
        if name is NOTSET:
            name = self._name
        return self.__class__(ns, plugin, name)

    def __str__(self) -> str:
        if self._name:
            return f'{self._ns}/{self._plugin}:{self._name}'
        return f'{self._ns}/{self._plugin}'

    @property
    def ns(self) -> str:
        return self._ns

    @property
    def plugin(self) -> str:
        return self._plugin

    @property
    def name(self) -> Optional[str]:
        return self._name

    @property
    def plugin_import_path(self) -> str:
        ns = _escape_import_part(self._ns)
        plugin = _escape_import_part(self._plugin)
        return f'{PLUGINS_PACKAGE}.{ns}.{plugin}'



import importlib
import itertools
import os
import os.path
import pkgutil

from dl_plus.const import PLUGINS_PACKAGE
from dl_plus.exceptions import DLPlusException

from .peqn import PEQN


class ExtractorLoadError(DLPlusException):

    pass


def discover_extractor_plugins_gen():
    try:
        extractors_package = importlib.import_module(PLUGINS_PACKAGE)
    except ImportError:
        return
    for ns_path in extractors_package.__path__:
        if not os.path.isdir(ns_path):
            continue
        for entry in os.scandir(ns_path):
            if not entry.is_dir():
                continue
            ns = entry.name
            for _, name, _ in pkgutil.iter_modules([entry.path]):
                yield f'{PLUGINS_PACKAGE}.{ns}.{name}'


def load_extractors_by_plugin_import_path(plugin_import_path, names=None):
    try:
        plugin_module = importlib.import_module(plugin_import_path)
    except ImportError as exc:
        raise ExtractorLoadError(f'cannot import {plugin_import_path}: {exc}')
    try:
        plugin = plugin_module.plugin
    except AttributeError:
        raise ExtractorLoadError(
            f'{plugin_import_path} does not marked as an extractor plugin')
    if names:
        try:
            extractors = list(map(plugin.get_extractor, names))
        except KeyError as exc:
            raise ExtractorLoadError(
                f'{plugin_import_path} does not contain {exc.args[0]}')
    else:
        extractors = plugin.get_all_extractors()
    return extractors


def load_extractors_by_peqn(peqn):
    if not isinstance(peqn, PEQN):
        try:
            peqn = PEQN.from_string(peqn)
        except ValueError:
            raise ExtractorLoadError(f'failed to parse PEQN: {peqn}')
    if peqn.name:
        names = [peqn.name]
    else:
        names = None
    return load_extractors_by_plugin_import_path(
        peqn.plugin_import_path, names)


def load_all_extractors():
    return list(itertools.chain.from_iterable(
        map(
            load_extractors_by_plugin_import_path,
            discover_extractor_plugins_gen(),
        )
    ))



import hashlib
import json
from collections import namedtuple
from io import BytesIO
from typing import Dict, Optional
from urllib.error import HTTPError
from urllib.request import urlopen

from dl_plus.exceptions import DLPlusException


Wheel = namedtuple('Wheel', 'name,version,metadata,filename,url,sha256')


class PyPIClientError(DLPlusException):

    pass


class RequestError(PyPIClientError):

    pass


class ParseError(PyPIClientError):

    pass


class DownloadError(RequestError):

    def __init__(
        self, project_name: str, version: Optional[str], error: str,
    ) -> None:
        if version:
            message = f'{project_name} {version}: {error}'
        else:
            message = f'{project_name}: {error}'
        super().__init__(message)


class Metadata(dict):

    @property
    def name(self) -> str:
        return self['info']['name']

    @property
    def version(self) -> str:
        return self['info']['version']

    @property
    def urls(self) -> Dict[str, Dict]:
        return self['urls']


class PyPIClient:

    JSON_BASE_URL = 'https://pypi.org/pypi'

    def build_json_url(
        self, project_name: str, version: Optional[str] = None,
    ) -> str:
        parts = [self.JSON_BASE_URL, project_name]
        if version:
            parts.append(version)
        parts.append('json')
        return '/'.join(parts)

    def fetch_metadata(
        self, project_name: str, version: Optional[str] = None,
    ) -> Metadata:
        url = self.build_json_url(project_name, version)
        try:
            with urlopen(url) as response:
                return Metadata(json.load(response))
        except (OSError, ValueError) as exc:
            raise RequestError from exc

    def _is_wheel_release(self, release: Dict) -> bool:
        return (
            release['packagetype'] == 'bdist_wheel' and not release['yanked'])

    def fetch_wheel_info(
        self, project_name: str, version: Optional[str] = None,
    ) -> Wheel:
        try:
            metadata = self.fetch_metadata(project_name, version)
        except RequestError as exc:
            if isinstance(exc.error, HTTPError) and exc.error.code == 404:
                error = 'not found'
            else:
                error = 'unexpected error'
            raise DownloadError(project_name, version, error) from exc
        try:
            release = next(filter(self._is_wheel_release, metadata.urls))
        except StopIteration:
            raise DownloadError(project_name, version, 'no wheel distribution')
        return Wheel(
            name=metadata.name,
            version=metadata.version,
            metadata=metadata,
            filename=release['filename'],
            url=release['url'],
            sha256=release['digests']['sha256'],
        )

    def download_file(self, url: str, sha256: Optional[str] = None) -> BytesIO:
        try:
            with urlopen(url) as response:
                buffer = BytesIO(response.read())
        except OSError as exc:
            raise DownloadError(f'{url}: {exc}') from exc
        if sha256:
            digest = hashlib.sha256(buffer.getvalue())
            hexdigest = digest.hexdigest()
            if hexdigest != sha256:
                raise DownloadError(
                    f'{url}: sha256 mismatch: expected {sha256}, '
                    f'got {hexdigest}'
                )
        return buffer



from enum import Enum


class NotSet(Enum):
    # Am I the only one who finds this hideous?
    # https://github.com/python/typing/issues/236
    # https://www.python.org/dev/peps/pep-0484/#support-for-singleton-types\
    # -in-unions
    NOTSET = 0


NOTSET = NotSet.NOTSET



import importlib
import sys
from io import StringIO

from .exceptions import DLPlusException


class YoutubeDLError(DLPlusException):

    pass


class UnknownBuiltinExtractor(YoutubeDLError):

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return f'unknown built-in extractor: {self.name}'


_NAME_PART_SURROGATE = '_'


_ytdl_module = None
_ytdl_module_name = None

_extractors = None
_extractors_registry = None


def _check_initialized():
    global _ytdl_module
    if not _ytdl_module:
        raise YoutubeDLError('not initialized')


def init(ytdl_module_name: str) -> None:
    global _ytdl_module
    if _ytdl_module:
        raise YoutubeDLError('already initialized')
    try:
        _ytdl_module = importlib.import_module(ytdl_module_name)
    except ImportError as exc:
        raise YoutubeDLError(f'failed to initialize: {exc}') from exc
    global _ytdl_module_name
    _ytdl_module_name = ytdl_module_name


def run(args):
    _check_initialized()
    global _ytdl_module_name
    orig_sys_argv = sys.argv
    try:
        sys.argv = [_ytdl_module_name.replace('_', '-'), *args]
        _ytdl_module.main()
    finally:
        sys.argv = orig_sys_argv


def get_ytdl_module():
    _check_initialized()
    global _ytdl_module
    return _ytdl_module


def get_help():
    _check_initialized()
    with StringIO() as buffer:
        stdout, stderr = sys.stdout, sys.stderr
        sys.stdout = sys.stderr = buffer
        try:
            _ytdl_module.main(['--help'])
        except SystemExit:
            pass
        finally:
            sys.stdout, sys.stderr = stdout, stderr
        return buffer.getvalue().partition('Options:')[2]


def import_module(module_name):
    _check_initialized()
    global _ytdl_module_name
    return importlib.import_module(f'{_ytdl_module_name}.{module_name}')


def _import_from(module, name):
    try:
        return getattr(module, name)
    except Exception as exc:
        raise ImportError(f'failed to import {name}: {exc}') from exc


def import_from(module_name, names):
    module = import_module(module_name)
    if isinstance(names, str):
        return _import_from(module, names)
    return tuple(_import_from(module, name) for name in names)


def get_all_extractors(*, include_generic: bool):
    global _extractors
    if _extractors is None:
        extractor_module = import_module('extractor')
        _extractors = tuple(extractor_module._ALL_CLASSES)
    if include_generic:
        return _extractors
    return _extractors[:-1]


def _get_extractor_name(extractor):
    ie_name = extractor.IE_NAME
    if isinstance(ie_name, property):
        return extractor().IE_NAME
    return ie_name


def _build_extractors_registry():
    registry = {}
    for extractor in get_all_extractors(include_generic=True):
        name_parts = _get_extractor_name(extractor).split(':')
        name_parts.reverse()
        _store_extractor_in_registry(extractor, name_parts, registry)
    return registry


def _store_extractor_in_registry(extractor, name_parts, registry):
    name_part = name_parts.pop()
    if name_part in registry:
        stored = registry[name_part]
        if not isinstance(stored, dict):
            subregistry = registry[name_part] = {_NAME_PART_SURROGATE: stored}
        else:
            subregistry = stored
        if not name_parts:
            if _NAME_PART_SURROGATE in subregistry:
                raise YoutubeDLError(f'duplicate name: {extractor!r}')
            subregistry[_NAME_PART_SURROGATE] = extractor
        else:
            _store_extractor_in_registry(extractor, name_parts, subregistry)
    elif not name_parts:
        registry[name_part] = extractor
    else:
        subregistry = registry[name_part] = {}
        _store_extractor_in_registry(extractor, name_parts, subregistry)


def _flatten_registry_gen(registry):
    for value in registry.values():
        if not isinstance(value, dict):
            yield value
        else:
            yield from _flatten_registry_gen(value)


def _get_extractors_from_registry(name_parts, registry):
    name_part = name_parts.pop()
    stored = registry[name_part]
    if not name_parts:
        if not isinstance(stored, dict):
            return [stored]
        return list(_flatten_registry_gen(stored))
    if not isinstance(stored, dict):
        raise KeyError(name_part)
    return _get_extractors_from_registry(name_parts, stored)


def get_extractors_by_name(name):
    global _extractors_registry
    if _extractors_registry is None:
        _extractors_registry = _build_extractors_registry()
    name_parts = name.split(':')
    name_parts.reverse()
    try:
        return _get_extractors_from_registry(name_parts, _extractors_registry)
    except KeyError:
        raise UnknownBuiltinExtractor(name)


def patch_extractors(extractors):
    extractor_module = import_module('extractor')
    extractor_module.gen_extractor_classes = lambda: extractors
    ie_keys_extractors_map = {
        extractor.ie_key(): extractor for extractor in extractors}
    extractor_module.get_info_extractor = (
        lambda ie_key: ie_keys_extractors_map[ie_key])
    importlib.reload(import_module('YoutubeDL'))



import json
import sys
from collections import namedtuple
from pathlib import Path

from dl_plus import ytdl
from dl_plus.config import get_config_dir_path
from dl_plus.exceptions import DLPlusException
from dl_plus.pypi import Metadata


backends_dir = get_config_dir_path() / 'backends'


BackendInfo = namedtuple(
    'BackendInfo', 'import_name,version,path,is_managed,metadata')


class BackendError(DLPlusException):

    pass


def _is_managed(location: Path) -> bool:
    try:
        location.relative_to(backends_dir)
        return True
    except ValueError:
        return False


def _normalize(string: str) -> str:
    return string.replace('-', '_')


def get_backend_dir(backend: str) -> Path:
    return backends_dir / _normalize(backend)


def parse_backend_string(backend_string: str):
    if '/' in backend_string:
        backend, _, package_name = backend_string.partition('/')
        backend_dir = get_backend_dir(backend)
        if not backend_dir.is_dir():
            raise BackendError(
                f'{backend_dir} does not exist or is not a directory')
    else:
        package_name = backend_string
        backend_dir = get_backend_dir(backend_string)
        if not backend_dir.is_dir():
            backend_dir = None
    return backend_dir, _normalize(package_name)


def save_metadata(backend_dir: Path, metadata: Metadata) -> None:
    with open(backend_dir / 'metadata.json', 'w') as fobj:
        json.dump(metadata, fobj)


def load_metadata(backend_dir: Path) -> Metadata:
    try:
        with open(backend_dir / 'metadata.json') as fobj:
            return Metadata(json.load(fobj))
    except OSError:
        return None


def init_backend(backend_string: str) -> BackendInfo:
    backend_dir, package_name = parse_backend_string(backend_string)
    if backend_dir:
        sys.path.insert(0, str(backend_dir))
    ytdl.init(package_name)
    ytdl_module = ytdl.get_ytdl_module()
    path = Path(ytdl_module.__path__[0])
    is_managed = _is_managed(path)
    if is_managed:
        metadata = load_metadata(backend_dir)
    else:
        metadata = None
    return BackendInfo(
        import_name=ytdl_module.__name__,
        version=ytdl.import_from('version', '__version__'),
        path=path,
        is_managed=is_managed,
        metadata=metadata,
    )



from typing import TYPE_CHECKING, Dict, Iterable, List, Type

from . import ytdl
from .extractor import machinery


if TYPE_CHECKING:
    from .extractor.extractor import Extractor


def get_extractors(names: Iterable[str]) -> List[Type['Extractor']]:
    extractors_dict: Dict[Type['Extractor'], bool] = {}
    for name in names:
        if name == ':builtins:':
            extractors = ytdl.get_all_extractors(include_generic=False)
        elif name == ':plugins:':
            extractors = machinery.load_all_extractors()
        elif '/' in name:
            extractors = machinery.load_extractors_by_peqn(name)
        else:
            extractors = ytdl.get_extractors_by_name(name)
        for extractor in extractors:
            if extractor not in extractors_dict:
                extractors_dict[extractor] = True
    return list(extractors_dict.keys())


def enable_extractors(names) -> None:
    extractors = get_extractors(names)
    ytdl.patch_extractors(extractors)



import itertools
import os
import shlex
from configparser import ConfigParser
from pathlib import Path
from typing import List, Optional, Union

from .exceptions import DLPlusException


DEFAULT_CONFIG = """
[main]
backend = youtube_dl

[extractors.enable]
:builtins:
:plugins:
generic
"""


class ConfigError(DLPlusException):

    pass


_config_dir_path: Optional[Path] = None


def get_config_dir_path() -> Path:
    global _config_dir_path
    if _config_dir_path:
        return _config_dir_path
    if os.name == 'nt':
        app_data = os.getenv('AppData')
        if app_data:
            parent = Path(app_data)
        else:
            parent = Path.home() / 'AppData' / 'Roaming'
    else:
        xdg_config_home = os.getenv('XDG_CONFIG_HOME')
        if xdg_config_home:
            parent = Path(xdg_config_home)
        else:
            parent = Path.home() / '.config'
    path = parent / 'dl-plus'
    _config_dir_path = path
    return path


class _Config(ConfigParser):

    def __init__(self) -> None:
        super().__init__(
            allow_no_value=True,
            delimiters=('=',),
            comment_prefixes=('#', ';'),
            inline_comment_prefixes=None,
            strict=True,
            empty_lines_in_values=False,
            default_section=None,
            interpolation=None,
        )


class Option:

    __slots__ = ('section', 'option')

    def __init__(self, section: str, option: str) -> None:
        self.section = section
        self.option = option

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return instance.get(self.section, self.option)


class Config(_Config):

    _UPDATE_SECTIONS = ('main',)
    _REPLACE_SECTIONS = ('extractors.enable', 'backend-options')

    def __init__(self) -> None:
        super().__init__()
        self.read_string(DEFAULT_CONFIG)

    def load(self, path: Union[Path, str, None] = None) -> None:
        if not path:
            path = get_config_dir_path() / 'config.ini'
            if not path.is_file():
                return
        else:
            if isinstance(path, str):
                path = Path(path)
            if not path.is_file():
                raise ConfigError(
                    f'failed to load config: {path} is not a file')
        config = _Config()
        try:
            with open(path) as fobj:
                config.read_file(fobj)
        except OSError as exc:
            raise ConfigError(f'failed to load config: {exc}') from exc
        for section in self._UPDATE_SECTIONS:
            self._load_section(section, config, replace=False)
        for section in self._REPLACE_SECTIONS:
            self._load_section(section, config, replace=True)

    def _load_section(self, name: str, config: _Config, replace: bool) -> None:
        if not config.has_section(name):
            return
        if not self.has_section(name):
            self.add_section(name)
        section = self[name]
        if replace:
            section.clear()
        section.update(config[name])

    def get_backend_options(self) -> Optional[List[str]]:
        if 'backend-options' not in self:
            return None
        return list(itertools.chain.from_iterable(
            map(shlex.split, self.options('backend-options'))))

    backend = Option('main', 'backend')



import re
from typing import Match, Optional

from dl_plus import ytdl


InfoExtractor = ytdl.import_from('extractor.common', 'InfoExtractor')
ExtractorError = ytdl.import_from('utils', 'ExtractorError')


class Extractor(InfoExtractor):
    """
    A base class for pluggable extractors
    """

    # Set by `ExtractorPlugin.register`, do not override.
    IE_NAME: Optional[str] = None

    @classmethod
    def ie_key(cls):
        return cls.IE_NAME

    # dl-plus extra attributes/methods

    DLP_BASE_URL: Optional[str] = None
    DLP_REL_URL: Optional[str] = None

    @classmethod
    def dlp_match(cls, url: str) -> Optional[Match[str]]:
        # a copy/paste from youtube-dl
        if '_VALID_URL_RE' not in cls.__dict__:
            cls._VALID_URL_RE = re.compile(cls._VALID_URL)
        return cls._VALID_URL_RE.match(url)



import sys

from .plugin import ExtractorPlugin


_LAZY_LOAD_NAMES = [
    'Extractor',
    'ExtractorError',
]

__all__ = _LAZY_LOAD_NAMES + ['ExtractorPlugin']


def __getattr__(name):
    if name not in _LAZY_LOAD_NAMES:
        raise AttributeError(name)
    from . import extractor
    _globals = globals()
    for _name in _LAZY_LOAD_NAMES:
        _globals[_name] = getattr(extractor, _name)
    return _globals[name]


# Python 3.6 — emulate PEP 562
# https://docs.python.org/3/reference/\
#   datamodel.html#customizing-module-attribute-access
if sys.version_info < (3, 7):
    from types import ModuleType
    sys.modules[__name__].__class__ = type(
        'LazyModule', (ModuleType,),
        {'__getattr__': staticmethod(__getattr__)},
    )
